from datetime import datetime, timedelta
from unittest.mock import MagicMock, patch

import pytest
from advanced_alchemy.exceptions import NotFoundError
from sqlalchemy import delete
from sqlalchemy.exc import NoResultFound

from leaguemanager import LeagueManager
from leaguemanager import models as m
from leaguemanager import services as s
from leaguemanager.core.toolbox import clear_table
from leaguemanager.schema import Day, GameTime, Month, Year
from leaguemanager.services import RoundRobinSchedule

org_data = {
    "name": "Test Organization",
}

league_data = {
    "name": "Test League",
}


season_data = {
    "name": "Season Test",
    "description": "Burndt",
    "projected_start_date": "2022-03-05:13:00:00",
}


team_data = [
    {"name": "Loro"},
    {"name": "June"},
    {"name": "Tripoli"},
    {"name": "Tres"},
]

ruleset_data = {
    "game_length": 45,
    "time_between": 15,
    "number_of_teams": 4,
    "number_of_phases": 10,
    "venue_count": 1,
}


@pytest.fixture(scope="module")
def _setup_db(
    org_service,
    league_service,
    season_service,
    team_service,
    team_member_service,
    ruleset_service,
    session,
):
    """Fixture to set up the database with test data."""
    # Clear existing data
    clear_table(session, m.Organization)
    clear_table(session, m.League)
    clear_table(session, m.Season)
    clear_table(session, m.Team)
    clear_table(session, m.Ruleset)

    # Create organization
    org = org_service.create(org_data, auto_commit=True)

    # Create League linked to the organization
    league_data["organization_id"] = org.id
    league = league_service.create(league_data, auto_commit=True)

    # Create Season linked to the league
    season_data["league_id"] = league.id
    season = season_service.create(season_data, auto_commit=True)

    # Create TeamMembership objects for each team
    for team in team_data:
        team_membership = team_member_service.create(
            {
                "season_id": season.id,
                "label": f"{team['name']} - Season: {season.name}",
                "auto_commit": True,
            }
        )
        team["team_membership_id"] = team_membership.id
        team_service.create(team, auto_commit=True)

    # Create Ruleset linked to the season
    ruleset_data["season_id"] = season.id
    _ = ruleset_service.create(ruleset_data, auto_commit=True)


@pytest.fixture(scope="module")
def season(_setup_db, season_service):
    """Fixture to provide the current season."""
    try:
        return season_service.get(id_attribute="name", item_id=season_data["name"])
    except NotFoundError as e:
        raise ValueError(f"Season with name {season_data['name']} does not exist in the database.") from e


@pytest.fixture(scope="module")
def scheduler(
    season,
    season_service,
    ruleset_service,
    phase_service,
    fixture_service,
    team_service,
):
    """Fixture to provide a RoundRobinSchedule instance."""
    mock_lm = MagicMock(spec=LeagueManager)
    scheduler = RoundRobinSchedule(mock_lm, season)

    # Patch the season_service and ruleset_service into RoundRobinSchedule properties
    mock_lm.provide_db_service.side_effect = lambda service_type: {
        s.SeasonService: season_service,
        s.RulesetService: ruleset_service,
        s.PhaseService: phase_service,
        s.FixtureService: fixture_service,
        s.TeamService: team_service,
    }.get(service_type, None)

    return scheduler


def test_round_robin_schedule_initialization(scheduler, season):
    """Test the initialization of the RoundRobinSchedule service."""

    assert isinstance(scheduler, RoundRobinSchedule)
    assert scheduler.season == season
    assert len(scheduler.teams) == 4


def test_round_robin_split_teams(scheduler):
    """Test the splitting of teams into fixtures."""
    teams_a, teams_b = scheduler.split_teams(scheduler.teams)

    assert len(teams_a) == 2
    assert len(teams_b) == 2


@pytest.mark.parametrize(
    "venue_count, match_number, expected_venue",
    [
        (1, 1, 1),
        (1, 2, 1),
        (3, 3, 3),
        (3, 4, 1),
        (2, 2, 2),
        (2, 3, 1),
        (2, 6, 2),
        (4, 10, 2),
    ],
)
def test_round_robin_generate_venue(scheduler, venue_count, match_number, expected_venue):
    """Test the generation of venues for the round-robin schedule."""
    scheduler.ruleset.venue_count = venue_count
    venue = scheduler.generate_venue_number(match_number)
    assert venue == expected_venue


@pytest.mark.parametrize(
    "td_minutes, multiplier, expected_hrs, expected_mins",
    [
        (35, 1, 0, 35),
        (75, 1, 1, 15),
        (125, 1, 2, 5),
        (60, 1, 1, 0),
        (45, 2, 1, 30),
        (35, 5, 2, 55),
    ],
)
def test_round_robin_convert_to_hrs_mins(scheduler, td_minutes, expected_hrs, expected_mins, multiplier):
    td = timedelta(minutes=td_minutes)
    hours, minutes = scheduler._convert_to_hours_and_minutes(td, multiplier)
    assert hours == expected_hrs
    assert minutes == expected_mins


@pytest.mark.parametrize(
    "game_length, time_between, match_count, venue_count, expected_hour, expected_minute",
    [
        (45, 15, 1, 1, 13, 0),
        (30, 10, 2, 1, 13, 40),
        (60, 0, 2, 1, 14, 0),
        (45, 20, 4, 2, 14, 5),
        (40, 5, 11, 3, 15, 15),
        (50, 10, 6, 2, 15, 0),
        (25, 0, 9, 1, 16, 20),  # 25 mins * 8 = 200 mins, 3 hrs 20 mins
    ],
)
def test_round_robin_increment_time(
    scheduler,
    game_length,
    time_between,
    venue_count,
    match_count,
    expected_hour,
    expected_minute,
):
    """Test the time increment logic in the round-robin schedule."""
    scheduler.ruleset.game_length = game_length
    scheduler.ruleset.time_between = time_between
    scheduler.ruleset.venue_count = venue_count
    match_count = match_count

    match_date = scheduler.season.projected_start_date

    gametime = scheduler._increment_time(match_count, match_date)

    expected = GameTime(
        year=Year(match_date.year),
        month=Month(match_date.month),
        day=Day(match_date.day),
        hour=expected_hour,
        minute=expected_minute,
        total_duration_in_minutes=scheduler.ruleset.game_length + scheduler.ruleset.time_between,
    )

    assert expected == gametime


def test_round_robin_get_or_create_phases(
    scheduler,
    season,
    season_service,
    ruleset_service,
    phase_service,
):
    """Test the get_or_create_phases method."""
    phases = scheduler.season.phases
    assert len(phases) == 0  # Initially, there should be no phases

    _ = scheduler.get_or_create_phases("GameWeek")
    phases = scheduler.season.phases
    assert len(phases) == 10  # Assuming 10 phases as per ruleset


def test_round_robin_create_matchups(scheduler):
    """Test the create_matchups method."""
    ...



from __future__ import annotations

from typing import Iterator

from attrs import define, field
from sqlalchemy import select
from sqlalchemy.orm import Session

from leaguemanager.services._typing import (
    AsyncRepositoryT,
    AsyncServiceT,
    ImporterT,
    ModelT,
    ScheduleServiceT,
    SyncRepositoryT,
    SyncServiceT,
)

__all__ = ["ServiceManagement", "RepositoryManagement", "ImporterManagement"]


@define
class ServiceManagement:
    """Manages a SQLAlchemySyncRepositoryService[ModelT] class.

    TODO: Async support

    Given the `service_type` and `db_session`, as well as a db_session, it will hold then provide
    the applicable service (for the corresponding service type.). The `get_service` property will
    return the appropriate service for the given `service_type` and `db_session`.

    Attributes:
        service_type (type[SyncServiceT] | type[AsyncServiceT]): Service type to manage.
        model_type (type[ModelT]): Model type for the given `service_type`.
        db_session (Session | None): Database session to use for the service.

    Example:
      >>> _service = ServiceManagement(
      ...     service_type=SeasonSyncService, model_type=Season, db_session=session
      ... )
      >>> _service.get_service
    """

    service_type: type[SyncServiceT] = field()
    db_session: Session = field(default=None)
    model_type: type[ModelT] = field(init=False)

    def __attrs_post_init__(self):
        self.model_type = self.service_type.repository_type.model_type

    @property
    def get_service(self) -> Iterator[ServiceManagement.service_type]:  # type: ignore[return]
        with self.service_type.new(session=self.db_session, statement=select(self.model_type)) as service:
            yield service


@define
class RepositoryManagement:
    """Manages a SQLAlchemySyncRepository class.

    TODO: Async support

    Given the `repository_type` as well as a db_session, it will hold then provide
    the applicable repository (for the corresponding repository type.). The `get_repository` property will
    return the appropriate repository for the given `repository_type` and `db_session`.

    Attributes:
        repository_type (type[SyncRepositoryT] | type[AsyncRepositoryT]): Repository type to manage.
        db_session (Session | None): Database session to use for the repository.

    Example:
      >>> _repository = RepositoryManagement(
      ...     repository_type=SeasonSyncRepository, db_session=session
      ... )
      >>> _repository.get_repository
    """

    repository_type: type[SyncRepositoryT] = field()
    db_session: Session = field(default=None)

    @property
    def get_repository(self) -> Iterator[RepositoryManagement.repository_type]:  # type: ignore[return]
        return self.repository_type(session=self.db_session, statement=select(self.repository_type.model_type))


@define
class ImporterManagement:
    """Manages Importer services.

    Provides a basic utility to manage importers for league data. It takes an `importer_type` and provides
    an instance of that importer type through the `get_importer` property. It can be used to register
    a specific importer type for league data management.

    Attributes:
        importer_type (type[ImporterT]): Importer type to manage.

    Example:
      >>> _importer = ImporterManagement(importer_type=LeagueImporter)
      >>> _importer.get_importer
    """

    importer_type: type[ImporterT] = field()

    @property
    def get_importer(self) -> ImporterT:
        return self.importer_type()


@define
class SchedulerManagement:
    """Manages scheduling services.

    Provides a basic utility to manage scheduling services for league data. It takes a `scheduler_type` and provides
    an instance of that scheduler type through the `get_scheduler` property. It can be used to register
    a specific scheduler type for league scheduling management.

    Attributes:
        service_type (type[SyncServiceT]): Service type to manage.

    Example:
      >>> _scheduler = SchedulerManagement(service_type=BracketSchedule)
      >>> _scheduler.get_service
    """

    scheduler_type: type[ScheduleServiceT] = field()

    @property
    def get_scheduler(self) -> ScheduleServiceT:
        return self.scheduler_type()



from __future__ import annotations

from typing import TYPE_CHECKING

from attr import define, field

from leaguemanager import models as m
from leaguemanager.schema import Day, GameTime, Month, Year

if TYPE_CHECKING:
    from leaguemanager.services import FixtureService, RulesetService, SeasonService

__all__ = ["RoundRobinPlayoffSchedule"]


@define
class RoundRobinPlayoffSchedule:
    """Service for generating and managing round-robin playoff schedules."""

    season_service: SeasonService = field(default=None)
    ruleset_service: RulesetService = field(default=None)
    fixture_service: FixtureService = field(default=None)

    def generate_schedule(self) -> None:
        """Generate the round-robin playoff schedule."""
        # Implementation for generating a round-robin playoff schedule
        raise NotImplementedError("Round-robin playoff schedule generation not implemented.")

    def get_schedule(self) -> None:
        """Get the generated round-robin playoff schedule."""
        # Implementation for retrieving the round-robin playoff schedule
        raise NotImplementedError("Retrieving round-robin playoff schedule not implemented.")

    def update_schedule(self) -> None:
        """Update the existing round-robin playoff schedule."""
        # Implementation for updating the round-robin playoff schedule
        raise NotImplementedError("Updating round-robin playoff schedule not implemented.")

    @property
    def season(self) -> m.Season:
        """Get the current season."""
        if self.season_service is None:
            raise ValueError("SeasonService is not set.")
        return self.season_service.get_current_season()



from pathlib import Path
from typing import Protocol, runtime_checkable

from attrs import define, field, validators

__all__ = [
    "ExcelLoader",
    "CSVLoader",
    "JSONLoader",
    "GoogleSheetsLoader",
    "MemoryLoader",
]


@runtime_checkable
class Importer(Protocol):
    """Protocol for importers that can load data into the database."""

    def load(self) -> None:
        """Load data into the database."""

    def validate(self) -> None:
        """Validate the data before loading it into the database."""


def convert_str_to_path(value: str | Path) -> Path:
    """Convert a string to a Path object."""
    if isinstance(value, str):
        return Path(value)
    return value


@define
class ExcelLoader:
    """A class to load data from Excel files into the database."""

    file_path: str | Path | None = field(default=None, converter=convert_str_to_path)

    def load(self, data: dict) -> None:
        """Load data from an Excel file into the database."""
        # Implementation for loading data from Excel
        raise NotImplementedError("Excel loading not implemented yet.")

    def validate(self, data: dict) -> None:
        """Validate the data before loading it into the database."""
        # Implementation for validating Excel data
        raise NotImplementedError("Excel loading not implemented yet.")


@define
class CSVLoader:
    """A class to load data from CSV files into the database."""

    file_path: str | Path | None = field(default=None, converter=convert_str_to_path)

    def load(self, data: dict) -> None:
        """Load data from a CSV file into the database."""
        # Implementation for loading data from CSV
        raise NotImplementedError("CSV loading not implemented yet.")

    def validate(self, data: dict) -> None:
        """Validate the data before loading it into the database."""
        # Implementation for validating CSV data
        raise NotImplementedError("CSV loading not implemented yet.")


@define
class JSONLoader:
    """A class to load data from JSON files into the database."""

    file_path: str | Path | None = field(default=None, converter=convert_str_to_path)

    def load(self, data: dict) -> None:
        """Load data from a JSON file into the database."""
        # Implementation for loading data from JSON
        raise NotImplementedError("JSON loading not implemented yet.")

    def validate(self, data: dict) -> None:
        """Validate the data before loading it into the database."""
        # Implementation for validating JSON data
        raise NotImplementedError("JSON loading not implemented yet.")


@define
class GoogleSheetsLoader:
    """A class to load data from Google Sheets into the database."""

    sheet_id: str | None = field(default=None)
    spreadsheet_name: str | None = field(default=None)
    """Name of the Google Sheets spreadsheet to load data from."""

    def load(self, data: dict) -> None:
        """Load data from a Google Sheet into the database."""
        # Implementation for loading data from Google Sheets
        raise NotImplementedError("Google Sheets loading not implemented yet.")

    def validate(self, data: dict) -> None:
        """Validate the data before loading it into the database."""
        # Implementation for validating Google Sheets data
        raise NotImplementedError("Google Sheets loading not implemented yet.")


@define
class MemoryLoader:
    """A class to load data from memory into the database."""

    data: dict | None = field(default=None)

    def load(self, data: dict) -> None:
        """Load data from memory into the database."""
        # Implementation for loading data from memory

        return data

    def validate(self, data: dict) -> None:
        """Validate the data before loading it into the database."""
        # Implementation for validating memory data
        raise NotImplementedError("Memory loading not implemented yet.")



from __future__ import annotations

from typing import TYPE_CHECKING

from attr import define, field

from leaguemanager import models as m

if TYPE_CHECKING:
    from leaguemanager.services import FixtureService, RulesetService, SeasonService

__all__ = ["BracketSchedule"]


@define
class BracketSchedule:
    """Service for generating and managing bracket schedules."""

    season_service: SeasonService = field(default=None)
    ruleset_service: RulesetService = field(default=None)
    fixture_service: FixtureService = field(default=None)

    def generate_schedule(self) -> None:
        """Generate the bracket schedule."""
        # Implementation for generating a bracket schedule
        raise NotImplementedError("Bracket schedule generation not implemented.")

    def get_schedule(self) -> None:
        """Get the generated bracket schedule."""
        # Implementation for retrieving the bracket schedule
        raise NotImplementedError("Retrieving bracket schedule not implemented.")

    def update_schedule(self) -> None:
        """Update the existing bracket schedule."""
        # Implementation for updating the bracket schedule
        raise NotImplementedError("Updating bracket schedule not implemented.")



import pytest

from leaguemanager import services as s
from leaguemanager.dependency.dependency_registry import LeagueManager


@pytest.fixture(scope="module")
def lm():
    """Fixture to provide a LeagueManager instance."""
    return LeagueManager()


def test_leaguemanager(lm):
    """Test that the LeagueManager instance is created correctly."""
    assert lm is not None
    assert isinstance(lm, LeagueManager)

    roundrobin = lm.provide_db_service(service_type=s.RoundRobinSchedule)

    assert roundrobin is not None
    assert isinstance(roundrobin, s.RoundRobinSchedule)



from __future__ import annotations

import datetime
from collections import deque
from itertools import chain
from random import shuffle
from typing import Any, Generator, Iterable

from advanced_alchemy.exceptions import NotFoundError
from attrs import define, field, validators
from sqlalchemy.exc import NoResultFound
from typing_extensions import TYPE_CHECKING

from leaguemanager import models as m
from leaguemanager import services as s
from leaguemanager.dependency.dependency_registry import LeagueManager

if TYPE_CHECKING:
    from uuid import UUID

    from leaguemanager.models import League, Ruleset, Team, TeamStats


# @define
# class TimeOfDay:
#     hour: int = field(default=0, validator=validators.in_([i for i in range(24)]))
#     minute: int = field(default=0, validator=validators.in_([i for i in range(60)]))
#     twelve_hr_format: str | None = field(
#         converter=lambda x: f"{x.hour % 12 or 12}:{x.minute:02d} {'AM' if x.hour < 12 else 'PM'}"
#     )


@define
class TeamObj:
    team_name: str | None = field(default=None)
    team_id: UUID | None = field(default=None)
    bye_week: bool = field(default=False)


@define
class PhaseObj:
    phase_name: str | None = field(default=None)
    phase_number: int | None = field(default=None)
    phase_length: int = field(default=1)  # Length in days


@define
class EventDays:
    mon: bool = field(default=False)
    tue: bool = field(default=False)
    wed: bool = field(default=False)
    thu: bool = field(default=False)
    fri: bool = field(default=False)
    sat: bool = field(default=False)
    sun: bool = field(default=False)


@define
class EventMatchup:
    home_team: TeamObj | None = field(default=None)
    away_team: TeamObj | None = field(default=None)
    phase_number: int | None = field(default=None)
    venue: str | None = field(default=None)
    start_date: datetime.date | None = field(default=None)


@define
class Scheduler:
    type: str = field(default="ROUNDROBIN")
    number_of_teams: int = field(default=0)
    teams: list[m.Team | TeamObj] = field(default=field(factory=list))
    number_of_phases: int = field(default=1)
    venue_count: int = field(default=1)
    start_date_and_time: datetime.date | None = field(default=None)
    game_length: int = field(default=60)
    time_between_games: int = field(default=15)
    event_days: EventDays = field(default=EventDays())

    @classmethod
    def from_season(cls, season: m.Season) -> Scheduler:
        """Create a Scheduler instance from a Season object.

        This presumes that the Season has a valid Ruleset associated with it, and that
        Teams have already been associated through TeamMembership model associations.
        """
        lm = LeagueManager()
        season_service = lm.provide_db_service(s.SeasonService)

        if not season.ruleset:
            raise NotFoundError("Ruleset not found for the given season.")

        r: m.Ruleset = season.ruleset
        teams = season_service.all_teams(season.id)

        if len(teams) == 0:
            raise NotFoundError("No teams found for the given season.")
        if r.number_of_teams != len(teams):
            raise ValueError(
                f"Number of teams in the ruleset ({r.number_of_teams}) does not match the number of teams in the season ({len(teams)})."
            )

        return cls(
            type=r.schedule_type,
            number_of_teams=r.number_of_teams or 0,
            teams=teams,
            number_of_phases=r.number_of_phases or 0,
            venue_count=r.venue_count or 1,
            start_date_and_time=season.projected_start_date,
            game_length=r.game_length or 60,
            time_between_games=r.time_between or 15,
            event_days=EventDays(
                mon=r.mon_fixtures,
                tue=r.tue_fixtures,
                wed=r.wed_fixtures,
                thu=r.thu_fixtures,
                fri=r.fri_fixtures,
                sat=r.sat_fixtures,
                sun=r.sun_fixtures,
            ),
        )

    @classmethod
    def from_ruleset(cls, ruleset: m.Ruleset) -> Scheduler:
        """Create a Scheduler instance from a Ruleset object."""

        return cls(
            type=ruleset.schedule_type,
            number_of_teams=ruleset.number_of_teams or 0,
            teams=[],
            number_of_phases=ruleset.number_of_phases or 1,
            venue_count=ruleset.venue_count or 1,
            start_date_and_time=ruleset.start_date,
            game_length=ruleset.game_length or 60,
            time_between_games=ruleset.time_between or 15,
            event_days=EventDays(
                mon=ruleset.mon_fixtures,
                tue=ruleset.tue_fixtures,
                wed=ruleset.wed_fixtures,
                thu=ruleset.thu_fixtures,
                fri=ruleset.fri_fixtures,
                sat=ruleset.sat_fixtures,
                sun=ruleset.sun_fixtures,
            ),
        )

    @property
    def team_objs(self) -> list[TeamObj]:
        """Return a list of TeamObj instances for each team in the league."""
        if not self.teams:
            return [TeamObj(team_name=f"Team {i + 1}") for i in range(self.number_of_teams)]
        team_objs = []
        for team in self.teams:
            if isinstance(team, m.Team):
                team_objs.append(TeamObj(team_name=team.name, team_id=team.id))
            elif isinstance(team, TeamObj):
                team_objs.append(team)
            else:
                raise TypeError(f"Unexpected type {type(team)} in teams list.")
        return team_objs

    def create_events(self) -> list[EventMatchup]:
        """Create events based on the scheduling type."""
        if self.type == "ROUNDROBIN":
            return self.generate_round_robin_events()
        else:
            raise ValueError(f"Unsupported schedule type: {self.type}")

    def generate_round_robin_events(self, shuffle_order: bool = True) -> list[EventMatchup]:
        """Create events for a round-robin schedule."""

        if len(self.team_objs) < 2:
            raise ValueError("Not enough teams to create a round-robin schedule.")

        # Calculate full number of rounds, and remaining phases (if any)
        rounds, remaining = divmod(self.number_of_phases, self.number_of_teams - 1)

        if rounds == 0:
            raise ValueError("Not enough phases to create a round-robin schedule.")

        all_events = []
        if shuffle_order:
            shuffle(self.team_objs)

        for phase in range(self.number_of_phases - remaining):
            phase = +1
            round_number = phase // (len(self.team_objs))
            round_number += 1
            phase_events = self.phase_matchups(phase, round_number)
            all_events.extend(phase_events)

        # Handle remaining phases
        if remaining > 0:
            # create events for the remaining phases with generic teams
            # TODO: Implement logic to handle remaining phases
            pass
        return all_events

    def phase_matchups(self, phase: int, round_number: int) -> Generator[EventMatchup, None, None]:
        """Create events/matchups for a specific phase.

        See: https://en.wikipedia.org/wiki/Round-robin_tournament#Circle_method for
        details on round-robin scheduling. Splitting teams creates "bye week" team to
        ensure even matchups, and the phase number is used to rotate the teams for
        each phase of the round-robin schedule.
        """
        team_objs = self.team_objs
        if len(team_objs) < 2:
            raise ValueError("Not enough teams to create events.")

        # Split teams into two halves
        home, away = self._split_teams(team_objs)

        # Create matchup by zipping home and away teams
        matchups = list(zip(home, away, strict=False))

        # Anchor one team and rotate the rest based on phase
        if not phase == 1:
            _teams = deque(chain(*matchups))
            anchor = _teams.popleft()
            _teams.rotate(phase)
            _teams.appendleft(anchor)
            home, away = self._split_teams([*_teams])
            matchups = list(zip(home, away, strict=False))

        _event_matchups = self._create_matchups(matchups, phase, round_number)

        event_matchups = self._schedule_events(_event_matchups, phase, round_number)

        return event_matchups

    def _create_matchups(
        self, matchups: Iterable[tuple[TeamObj, TeamObj]], phase: int, round_number: int
    ) -> Generator[EventMatchup, None, None]:
        """Create EventMatchup instances from the matchups."""

        for match_number, team_pair in enumerate(matchups, start=1):
            if round_number % 2 == 0:
                home, away = team_pair
            else:
                home, away = team_pair[::-1]

            yield EventMatchup(
                home_team=home,
                away_team=away,
                phase_number=phase,
                venue=f"Venue {match_number % self.venue_count + 1}",
            )

    def _schedule_events(self, matchups: Iterable[EventMatchup], phase: int, round_number: int) -> list[EventMatchup]:
        """Schedule events based on the matchups and the start date/time."""

        if not self.start_date_and_time:
            raise ValueError("Start date and time must be set to schedule events.")

        scheduled_events = []
        current_date = self.start_date_and_time

        for matchup in matchups:
            matchup.start_date = current_date.date()
            matchup.venue = f"Venue {matchup.venue or 1}"

            # Check if concurrent games are allowed
            # Schedule games at same time based on how many concurrent games are allowed
            if self.venue_count > 1:
                matchup.start_date = current_date + datetime.timedelta(minutes=(matchup.venue or 1) * self.game_length)
            else:
                matchup.start_date = current_date

            # Increment the time for the next event
            current_date += datetime.timedelta(minutes=self.game_length + self.time_between_games)

            # Check if the event day is valid
            if not self._is_valid_event_day(current_date):
                continue

            scheduled_events.append(matchup)

        return scheduled_events

    def _is_valid_event_day(self, date: datetime.date) -> bool:
        """Check if the given date is a valid event day based on the event_days settings."""
        if self.event_days.mon and date.weekday() == 0:
            return True
        if self.event_days.tue and date.weekday() == 1:
            return True
        if self.event_days.wed and date.weekday() == 2:
            return True
        if self.event_days.thu and date.weekday() == 3:
            return True
        if self.event_days.fri and date.weekday() == 4:
            return True
        if self.event_days.sat and date.weekday() == 5:
            return True
        if self.event_days.sun and date.weekday() == 6:
            return True
        return False

    def _split_teams(self, team_obj: Iterable[TeamObj]) -> tuple[list[TeamObj], list[TeamObj]]:
        """Split a list of teams in half, creating "bye week" team if number of teams is odd."""

        if len(team_obj) % 2 != 0:
            self.teams.append(TeamObj(bye_week=True))

        return (
            team_obj[: len(team_obj) // 2],
            team_obj[len(team_obj) // 2 :],
        )



from pathlib import Path

import pytest

from leaguemanager.dependency.loader import DynamicObjectLoader


@pytest.fixture(scope="module")
def loader() -> DynamicObjectLoader:
    """Fixture to provide a DynamicObjectLoader instance."""
    loader = DynamicObjectLoader(app_dir=Path("src/leaguemanager/services"))

    return loader


def test_get_importers(loader: DynamicObjectLoader) -> None:
    """Test that the loader can retrieve Importer classes."""
    importers = loader.get_importers()
    assert len(importers) == 5, "Expected 5 importers to be found"


def test_get_schedule_services(loader: DynamicObjectLoader) -> None:
    """Test that the loader can retrieve ScheduleServiceBase classes."""
    sched_services = loader.get_schedule_services()
    assert len(sched_services) == 4, "Expected 4 schedule services to be found"



import pytest
from advanced_alchemy.exceptions import NotFoundError
from sqlalchemy import delete
from sqlalchemy.exc import NoResultFound

from leaguemanager import LeagueManager
from leaguemanager import models as m
from leaguemanager import services as s
from leaguemanager.core.toolbox import clear_table
# from leaguemanager.domain.competition import Scheduler


# @pytest.fixture(scope="module")
# def base_season(all_data, org_service, league_service, season_service):
#     org = org_service.create(all_data["organization"], auto_commit=True)
#     league = league_service.create(all_data["leagues"][0], auto_commit=True)
#     season = season_service.create(all_data["seasons"][0], auto_commit=True)

#     league.organization_id = org.id
#     season.league_id = league.id

#     league_service.update(league, auto_commit=True)
#     season_service.update(season, auto_commit=True)

#     return season


# def test_season(base_season):
#     assert base_season.name == "Test Season"
#     assert base_season.league.name == "Test League"


# def test_scheduler_properties_no_ruleset(base_season):
#     scheduler = Scheduler(season=base_season)
#     assert scheduler.season == base_season
#     assert scheduler.league == base_season.league

#     with pytest.raises(NotFoundError, match="Ruleset not found"):
#         _ = scheduler.ruleset

#     with pytest.raises(NotFoundError, match="Ruleset not found"):
#         _ = scheduler.total_games

#     with pytest.raises(NotFoundError, match="Ruleset not found"):
#         _ = scheduler.start_date

#     with pytest.raises(NotFoundError, match="Ruleset not found"):
#         _ = scheduler.total_teams


# def test_scheduler_properties_with_ruleset(base_season, season_service):
#     ruleset = m.Ruleset(
#         number_of_games=10,
#         start_date="2025-01-05 08:00:00",
#         schedule_type="ROUNDROBIN",
#         season_id=base_season.id,
#     )

#     base_season.ruleset = ruleset
#     season_service.update(base_season, auto_commit=True)
#     scheduler = Scheduler(season=base_season)

#     assert base_season.ruleset == ruleset
#     assert scheduler.ruleset == ruleset
#     assert scheduler.total_games == 10
#     assert scheduler.start_date == ruleset.start_date


# def test_scheduler_all_teams(session, base_season, season_service, team_service, team_member_service):
#     team_a_member = m.TeamMembership(label="Member A", season_id=base_season.id)
#     team_b_member = m.TeamMembership(label="Member B", season_id=base_season.id)
#     team_c_member = m.TeamMembership(label="Member C", season_id=base_season.id)
#     team_d_member = m.TeamMembership(label="Member B", season_id=base_season.id)

#     team_member_service.create_many([team_a_member, team_b_member, team_c_member, team_d_member], auto_commit=True)

#     _teams = [
#         m.Team(name="Team A", team_membership_id=team_a_member.id),
#         m.Team(name="Team B", team_membership_id=team_b_member.id),
#         m.Team(name="Team C", team_membership_id=team_c_member.id),
#         m.Team(name="Team D", team_membership_id=team_d_member.id),
#     ]

#     teams = team_service.create_many(_teams, auto_commit=True)
#     member = team_member_service.get(team_a_member.id)

#     assert member is not None
#     assert team_a_member.id == teams[0].team_membership_id

#     all_teams = season_service.all_teams(base_season.id)
#     assert len(all_teams) == 4
#     assert all(team.name in [t.name for t in all_teams] for team in teams)
#     assert len(base_season.team_memberships) == 4

#     clear_table(session, m.Team)
#     clear_table(session, m.TeamMembership)


# def test_scheduler_create_dummy_teams_if_none(session, base_season, season_service, team_member_service, team_service):
#     ruleset = m.Ruleset(
#         number_of_games=10,
#         number_of_teams=5,
#         start_date="2025-01-05 08:00:00",
#         schedule_type="ROUNDROBIN",
#         season_id=base_season.id,
#     )

#     base_season.ruleset = ruleset
#     season_service.update(base_season, auto_commit=True)
#     scheduler = Scheduler(
#         season=base_season,
#         season_service=season_service,
#         team_member_service=team_member_service,
#         team_service=team_service,
#     )

#     # Since no teams currently present, the `total_teams` property creates placeholder teams
#     assert scheduler.total_teams == 5

#     teams = scheduler.teams
#     assert sorted(team.name for team in teams) == sorted(["Team 1", "Team 2", "Team 3", "Team 4", "Team 5"])

#     clear_table(session, m.Team)
#     clear_table(session, m.TeamMembership)


# def test_scheduler_from_season(base_season):
#     scheduler = Scheduler().from_season(base_season)

#     assert scheduler.season is base_season


# def test_scheduler_from_season_with_wrong_type():
#     with pytest.raises(TypeError, match="Expected a Season object"):
#         Scheduler.from_season("not_a_season")


# def test_scheduler_split_teams(session, base_season, season_service, team_service, team_member_service):
#     # Create teams for the season
#     team_a = m.Team(name="Team A")
#     team_b = m.Team(name="Team B")
#     team_c = m.Team(name="Team C")
#     team_d = m.Team(name="Team D")

#     teams = team_service.create_many([team_a, team_b, team_c, team_d], auto_commit=True)

#     scheduler = Scheduler(
#         season=base_season,
#     )

#     split_teams = scheduler.split_teams(teams)

#     assert len(split_teams) == 2
#     assert len(split_teams[0]) == 2
#     assert len(split_teams[1]) == 2

#     team_e = m.Team(name="Team E")
#     teams.append(team_service.create(team_e, auto_commit=True))

#     with pytest.raises(ValueError, match="List length must be even."):
#         scheduler.split_teams(teams)

#     clear_table(session, m.Team)


# @pytest.fixture(scope="module")
# def sunday_schedule(season, league):
#     schedule = Schedule(
#         league_id=league.id,
#         season_id=season.id,
#         name="Schedule for Test Scheduler",
#         total_games=10,
#         start_date="2025-1-05 08:00:00",
#         concurrent_games=2,
#     )

#     return schedule


# @pytest.fixture(scope="module")
# def scheduler(session, sunday_schedule) -> Scheduler:
#     scheduler = Scheduler(session=session, schedule=sunday_schedule)
#     return scheduler


# def test_generator_split_teams_even(scheduler, teams):
#     split_teams = scheduler.generator.split_teams(teams)

#     assert split_teams == (teams[:4], teams[4:])


# def test_generator_split_teams_odd(scheduler, teams):
#     temp_remove_team = teams.pop()

#     assert len(teams) == 7

#     with pytest.raises(ValueError):
#         _ = scheduler.generator.split_teams(teams)

#     teams.append(temp_remove_team)


# @pytest.mark.parametrize("matchday", [1, 2, 3])
# def test_generator_create_matchups_first_matchday(scheduler, matchday, teams):
#     matchups = scheduler.generator.create_matchups(matchday=matchday, teams=teams)
#     assert len(matchups) == 4


# def test_increment_matchday(scheduler):
#     matchday_1_start_time = scheduler.generator.increment_matchday(matchday=1)
#     matchday_2_start_time = scheduler.generator.increment_matchday(matchday=2)

#     assert matchday_1_start_time == datetime(2025, 1, 5, 8, 0)
#     assert matchday_2_start_time == datetime(2025, 1, 12, 8, 0)


# @pytest.mark.parametrize(
#     "count, concurrent, expected",
#     [(1, 2, "A"), (2, 2, "B"), (3, 4, "C"), (4, 4, "D"), (2, 1, "A"), (3, 1, "A")],
# )
# def test_determine_field(scheduler, count, concurrent, expected):
#     field = scheduler.generator.determine_field(count, concurrent)
#     assert field == expected


# def test_home_or_away_even_round(scheduler, all_teams):
#     fake_match = (all_teams[0], all_teams[1])

#     home, away = scheduler.generator.home_or_away(round_number=2, match=fake_match)
#     assert home == all_teams[0]
#     assert away == all_teams[1]


# def test_home_or_away_odd_round(scheduler, all_teams):
#     fake_match = (all_teams[0], all_teams[1])

#     home, away = scheduler.generator.home_or_away(round_number=1, match=fake_match)
#     assert home == all_teams[1]
#     assert away == all_teams[0]


# @pytest.mark.parametrize(
#     "_concurrent, _matchday, _match_count, expected",
#     [
#         (1, 1, 1, datetime(2025, 1, 5, 8, 0)),
#         (1, 1, 2, datetime(2025, 1, 5, 10, 0)),
#         (1, 1, 3, datetime(2025, 1, 5, 12, 0)),
#         (1, 1, 4, datetime(2025, 1, 5, 14, 0)),
#         (1, 1, 5, datetime(2025, 1, 5, 16, 0)),
#         (1, 2, 1, datetime(2025, 1, 12, 8, 0)),
#         (1, 2, 3, datetime(2025, 1, 12, 12, 0)),
#         (1, 2, 5, datetime(2025, 1, 12, 16, 0)),
#         (2, 1, 1, datetime(2025, 1, 5, 8, 0)),
#         (2, 1, 2, datetime(2025, 1, 5, 8, 0)),
#         (2, 1, 3, datetime(2025, 1, 5, 10, 0)),
#         (2, 1, 4, datetime(2025, 1, 5, 10, 0)),
#         (2, 1, 5, datetime(2025, 1, 5, 12, 0)),
#         (2, 1, 6, datetime(2025, 1, 5, 12, 0)),
#         (3, 2, 1, datetime(2025, 1, 12, 8, 0)),
#         (3, 3, 2, datetime(2025, 1, 19, 8, 0)),
#         (3, 4, 3, datetime(2025, 1, 26, 8, 0)),
#         (3, 5, 4, datetime(2025, 2, 2, 10, 0)),
#         (3, 6, 7, datetime(2025, 2, 9, 12, 0)),
#         (4, 1, 1, datetime(2025, 1, 5, 8, 0)),
#         (4, 1, 4, datetime(2025, 1, 5, 8, 0)),
#     ],
# )
# def test_determine_start_time_third_matchday(
#     scheduler, sunday_schedule, _concurrent, _matchday, _match_count, expected
# ):
#     sunday_schedule.concurrent_games = _concurrent
#     start_time = scheduler.generator.determine_start_time(matchday=_matchday, match_count=_match_count)
#     assert start_time == expected


# def test_create_matchday_fixtures(scheduler, teams):
#     fixtures = scheduler.generator.create_matchday_fixtures(matchday=1, round_number=1, teams=teams)
#     assert len(fixtures) == 4


# @pytest.mark.parametrize(["total_games", "expected"], [(8, 32), (10, 32), (15, 32), (16, 64), (20, 64), (24, 96)])
# def test_generate_fixtures_eight_teams(scheduler, total_games, expected, teams):
#     scheduler.generator.schedule.total_games = total_games
#     fixtures = scheduler.generator.generate_fixtures(teams=teams)
#     assert len(fixtures) == expected

#     fixtures = scheduler.generator.generate_fixtures(shuffle_order=False, teams=teams)
#     assert len(fixtures) == expected
#     fixtures.clear()


# @pytest.mark.parametrize(["total_games", "expected"], [(4, 8), (8, 16), (10, 16)])
# def test_generate_fixtures_four_teams(scheduler, total_games, expected, teams):
#     teams = teams[:4]

#     scheduler.generator.schedule.total_games = total_games
#     fixtures = scheduler.generator.generate_fixtures(teams=teams)
#     assert len(fixtures) == expected

#     fixtures = scheduler.generator.generate_fixtures(shuffle_order=False, teams=teams)
#     assert len(fixtures) == expected
#     fixtures.clear()


# def test_generate_season_fixtures_too_few_teams(scheduler, session, teams):
#     scheduler.generator.schedule.total_games = 3
#     with pytest.raises(ValueError):
#         scheduler.generate_fixtures(teams=teams)
#     scheduler.generator.schedule.total_games = 10


# def test_sort_fixtures_by_date(session, scheduler, season, league, teams):
#     _schedule = Schedule(**schedule_data, season_id=season.id, league_id=league.id)
#     scheduler = Scheduler(session=session, schedule=_schedule)

#     fixtures = scheduler.generate_fixtures(teams=teams)
#     _schedule.fixtures = fixtures

#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 10, 0)

#     sorted_fixtures = scheduler.sort_fixtures(order_by_field="date", sort_order="desc")

#     assert sorted_fixtures[0].date == datetime(2025, 2, 23, 10, 0)
#     assert sorted_fixtures[-1].date == datetime(2025, 1, 5, 8, 0)


# def test_push_fixture_to_end_of_schedule(session, scheduler, league, season, teams):
#     clear_table(session, Fixture)

#     _schedule = Schedule(
#         league_id=league.id,
#         season_id=season.id,
#         name="Test Schedule",
#         total_games=10,
#         start_date="2025-1-05 08:00:00",
#     )
#     scheduler = Scheduler(session=session, schedule=_schedule)
#     _fixtures = scheduler.generator.generate_fixtures(teams=teams)
#     _schedule.fixtures = _fixtures

#     fixtures = scheduler.schedule.fixtures

#     assert len(fixtures) == 32
#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 10, 0)

#     scheduler.push_fixture_to_end_of_schedule(fixtures[0])

#     assert len(fixtures) == 32
#     assert fixtures[0].date == datetime(2025, 3, 2, 8, 0)
#     assert fixtures[1].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 10, 0)


# def test_push_matchday_to_end_of_schedule(session, scheduler, season, league, teams):
#     _schedule = Schedule(
#         league_id=league.id,
#         season_id=season.id,
#         name="Test Schedule",
#         total_games=10,
#         start_date="2025-1-05 08:00:00",
#     )

#     fixtures = scheduler.generator.generate_fixtures(teams=teams[:4])
#     _schedule.fixtures = fixtures
#     scheduler = Scheduler(session=session, schedule=_schedule)

#     scheduler.sort_fixtures()

#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[1].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[2].date == datetime(2025, 1, 12, 8, 0)
#     assert fixtures[3].date == datetime(2025, 1, 12, 8, 0)
#     assert fixtures[4].date == datetime(2025, 1, 19, 8, 0)
#     assert fixtures[5].date == datetime(2025, 1, 19, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 8, 0)

#     scheduler.push_matchday_to_end_of_schedule(2)

#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[1].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[2].date == datetime(2025, 3, 2, 8, 0)
#     assert fixtures[3].date == datetime(2025, 3, 2, 8, 0)
#     assert fixtures[4].date == datetime(2025, 1, 19, 8, 0)
#     assert fixtures[5].date == datetime(2025, 1, 19, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 2, 23, 8, 0)


# def test_push_all_fixtures_by_one_week(session, scheduler, league, season, teams):
#     _schedule = Schedule(
#         league_id=league.id,
#         season_id=season.id,
#         name="Test Schedule",
#         total_games=10,
#         start_date="2025-1-05 08:00:00",
#     )

#     fixtures = scheduler.generator.generate_fixtures(teams=teams[:4])
#     _schedule.fixtures = fixtures
#     scheduler = Scheduler(session, schedule=_schedule)

#     fixture = fixtures[4]  # Contains matchday 3 (1/19/2025)
#     scheduler.push_all_fixtures_by_one_week(fixture)
#     assert fixtures[0].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[1].date == datetime(2025, 1, 5, 8, 0)
#     assert fixtures[2].date == datetime(2025, 1, 12, 8, 0)
#     assert fixtures[3].date == datetime(2025, 1, 12, 8, 0)
#     assert fixtures[4].date == datetime(2025, 1, 26, 8, 0)
#     assert fixtures[5].date == datetime(2025, 1, 26, 8, 0)
#     assert fixtures[6].date == datetime(2025, 2, 2, 8, 0)
#     assert fixtures[7].date == datetime(2025, 2, 2, 8, 0)
#     assert fixtures[-1].date == datetime(2025, 3, 2, 8, 0)



from importlib import import_module
from importlib.util import module_from_spec, spec_from_file_location
from inspect import isclass
from pathlib import Path
from types import ModuleType
from typing import Any, Generator, Iterable

from advanced_alchemy.config import SQLAlchemyAsyncConfig, SQLAlchemySyncConfig
from advanced_alchemy.repository import SQLAlchemyAsyncRepository, SQLAlchemySyncRepository
from advanced_alchemy.service import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService
from attrs import define, field

from leaguemanager.core import get_settings
from leaguemanager.services.scheduling.base import ScheduleServiceBase
from leaguemanager.services.template_loader.league_importer import Importer

get_settings.cache_clear()
settings = get_settings()


@define
class DynamicObjectLoader:
    app_dir: Path = field()
    root_dir: Path = field()
    search_dir: Path = field()
    local_only: bool = field(default=False)

    @app_dir.default
    def _app_dir(self):
        return settings.LEAGUE_MANAGER_APP_DIR

    @root_dir.default
    def _root_dir(self):
        return settings.LEAGUE_MANAGER_ROOT_DIR

    @search_dir.default
    def _search_dir(self):
        return settings.LEAGUE_MANAGER_APP_DIR

    @classmethod
    def local_app(cls, app_dir: Path | None = None, root_dir: Path | None = None, search_dir: Path | None = None):
        """Creates a DynamicObjectLoader for a user's application

        Args:
            app_dir (Path | None, optional): The path to user's application. Defaults to None.
            root_dir (Path | None, optional): The path to the root of the user's application. Defaults to None.
            search_dir (Path | None, optional): The path to search for modules. Defaults to None.

        Returns:
            DynamicObjectLoader: A configured DynamicObjectLoader with defaults for user's application
        """
        if not app_dir:
            app_dir = settings.APP_DIR
        if not root_dir:
            root_dir = settings.APP_ROOT
        if not search_dir:
            search_dir = app_dir
        return cls(app_dir=app_dir, root_dir=root_dir, search_dir=search_dir, local_only=True)

    def _is_sync_repo(self, item: Any):
        """Checks if item is SQLAlchemySyncRepository class or subclass"""
        return isclass(item) and issubclass(item, SQLAlchemySyncRepository)

    def _is_async_repo(self, item: Any):
        """Checks if item is SQLAlchemyAsyncRepository class or subclass"""
        return isclass(item) and issubclass(item, SQLAlchemyAsyncRepository)

    def _is_sync_service(self, item: Any):
        """Checks if item is SQLAlchemySyncRepositoryService class or subclass."""
        return isclass(item) and issubclass(item, SQLAlchemySyncRepositoryService)

    def _is_async_service(self, item: Any):
        """Checks if item is SQLAlchemyAsyncRepositoryService class or subclass."""
        return isclass(item) and issubclass(item, SQLAlchemyAsyncRepositoryService)

    def _is_sync_config(self, item: Any):
        """Checks if item is a SQLAlchemySyncConfig.

        Using `isclass(item)` doesn't work here. May revisit at a later time.
        """
        return item.__class__ == SQLAlchemySyncConfig

    def _is_async_config(self, item: Any):
        """Checks if item is a SQLAlchemyAsyncConfig.

        Using `isclass(item)` doesn't work here. May revisit at a later time.
        """
        return item.__class__ == SQLAlchemyAsyncConfig

    def _is_importer(self, item: Any):
        """Checks if item is a Protocol of the Importer class."""
        return isclass(item) and issubclass(item, Importer)

    def _is_schedule(self, item: Any):
        """Checks if item is a ScheduleServiceBase class or subclass."""
        return isclass(item) and issubclass(item, ScheduleServiceBase)

    def get_repositories(
        self, is_async: bool = False, **kwargs
    ) -> list[SQLAlchemySyncRepository | SQLAlchemyAsyncRepository]:
        """Returns all SQLAlchemySyncRepository or SQLAlchemyAsyncRepository classes dynamically.

        Args:
            is_async (bool, optional): If True, returns all SQLAlchemyAsyncRepository classes.
            Defaults to False.
        """
        if is_async:
            return self.load_objects(self.search_dir, self._is_async_repo)
        return self.load_objects(self.search_dir, self._is_sync_repo)

    def get_services(
        self, is_async: bool = False, **kwargs
    ) -> list[SQLAlchemySyncRepositoryService | SQLAlchemyAsyncRepositoryService]:
        """Returns all SQLAlchemySyncRepositoryService or SQLAlchemyAsyncRepositoryService classes dynamically.

        Args:
            is_async (bool, optional): If True, returns all SQLAlchemyAsyncRepositoryService classes.
            Defaults to False.
        """
        if is_async:
            return self.load_objects(self.search_dir, self._is_async_service)
        return self.load_objects(self.search_dir, self._is_sync_service)

    def get_configs(self, is_async: bool = False) -> list[SQLAlchemySyncConfig | SQLAlchemyAsyncConfig]:
        """Gets all SQLAlchemySyncConfig or SQLAlchemyAsyncConfig classes dynamically.

        If try_local_first is True, the search will start on the host application, and if a config
        item is found, it will load instead of league manager's default.

        Args:
            is_async (bool, optional): If True, returns all SQLAlchemyAsyncConfig classes.
            Defaults to False.
        """
        if is_async:
            return self.load_objects(self.search_dir, self._is_async_config, local_only=self.local_only)
        return self.load_objects(self.search_dir, self._is_sync_config, local_only=self.local_only)

    def get_importers(self) -> list[Importer]:
        """Returns all Importer classes dynamically.

        This will return all classes that implement the Importer protocol.
        """
        return self.load_objects(self.search_dir, self._is_importer, local_only=self.local_only)

    def get_schedule_services(self) -> list[ScheduleServiceBase]:
        """Returns all ScheduleServiceBase classes dynamically.

        This is used to load scheduling services that are defined in the user's application.

        For now, only implemented to look at internally defined modules.
        """

        return self.load_objects(self.search_dir, self._is_schedule, local_only=self.local_only)

    def load_objects(self, search_dir: Path, compare: callable, local_only: bool = False) -> Iterable[ModuleType]:
        """Looks through a list of objects and returns them based on a comparison function.

        If `local_only` is True, then it will only search the local directory for modules. Use this to
        override League Manager's default search location.

        Args:
            search_dir (Path): Directory name to search recursively
            compare (bool): Boolean comparison of all py files in `search_dir` directory
            local_only (bool): If True, only search the local directory for modules. Defaults to False.

        Returns:
            list: All modules that match the `compare` function.
        """
        items = []
        if local_only:
            _full_paths = self.full_paths(search_dir)
            items += self.collect_matching_objects(_full_paths, compare, is_dotted_path=False)
        else:
            _dotted_paths = self.dotted_paths(search_dir)
            items += self.collect_matching_objects(_dotted_paths, compare)
        return items

    def collect_matching_objects(
        self, modules: Iterable[str | Path], compare: callable, is_dotted_path: bool = True
    ) -> Iterable[ModuleType]:
        """Looks for all classes that match the `compare` function in a list of dotted paths.

        It expects a list of dotted paths to modules. It attempts to import each item, ignoring
        circular imports and modules that don't exist. If a module has an `__all__` attribute, it
        will only run the compare function on items in that attribute, and will only return unique
        items found.

        Args:
            modules (list[str]): List of dotted paths to modules
            compare (callable | bool): Boolean comparison of all py files in `search_dir` directory. If set
                to True, then all items in the module will be returned.
            is_dotted_path (bool, optional): Expects `modules` to be a list of dotted relative paths to modules (e.g. "app.app").
                If False, expects a list of full paths (e.g. /home/user/app/app.py). Defaults to True.

        Returns:
            list: All modules that match the `compare` function.
        """
        items = []
        for mod in modules:
            if "__init__" in str(mod):
                continue
            try:
                if not is_dotted_path:
                    _module = self.import_module_from_full_path(mod.stem, str(mod))
                else:
                    _module = import_module(mod)

            except AttributeError:
                # Ignore modules that create circular imports
                continue
            except ModuleNotFoundError:
                # Ignore modules that don't exist
                continue
            except Exception as e:
                print(f"Failed to import {mod} due to: {e}")
                return None
            if hasattr(_module, "__all__"):
                objs = [getattr(_module, obj) for obj in _module.__all__]
                if not objs:
                    print("No objects found")
                    continue
                items += [o for o in objs if compare(o) and o not in items]
        return items

    def dotted_paths(self, search_path: Path) -> Generator[str, None, None]:
        """Creates dotted paths to all .py files in `search_path` directory.

        Will exclude any files in directories in the `_exclude` list.

        Args:
            search_path (Path): Directory name to search recursively

        Returns:
            Generator[str]: Generator of dotted paths to all .py files in `search_path` directory
        """
        _exclude = [".venv", "venv", "tests", "test", "docs", "migrations", "__main__", "__init__"]
        for _file in search_path.rglob("**/*.py"):
            mod_path = str(_file.relative_to(self.root_dir)).replace("/", ".")
            if any(e in mod_path for e in _exclude):
                continue
            yield f"{mod_path[:-3]}"

    def full_paths(self, search_path: Path) -> Generator[Path, None, None]:
        """Creates full paths to all .py files in `search_path` directory.

        Will exclude any files in directories in the `_exclude` list.

        Args:
            search_path (Path): Directory name to search recursively

        Returns:
            Generator[Path]: Generator of full paths to all .py files in `search_path` directory
        """
        _exclude = ["tests", "test", "docs", "migrations", "__main__", "__init__"]
        for _file in search_path.rglob("**/*.py"):
            if any(e in str(_file) for e in _exclude):
                continue
            yield _file

    def import_module_from_full_path(self, module_name: str, full_path: str) -> Any:
        spec = spec_from_file_location(module_name, full_path)
        _module = module_from_spec(spec)
        spec.loader.exec_module(_module)
        return _module



from __future__ import annotations

import typer
from sqlalchemy.orm import Session

from leaguemanager.dependency import LeagueManager
from leaguemanager.services._typing import (
    ImporterT,
    SQLAlchemyAsyncConfigT,
    SQLAlchemySyncConfigT,
    SyncRepositoryT,
    SyncServiceT,
)

registry = LeagueManager()


def provide_manager_service(param: typer.CallbackParam) -> SyncServiceT | SyncRepositoryT:
    return registry.provide_db_service(service_type=param.type.func)


def provide_manager_repository(param: typer.CallbackParam) -> SyncRepositoryT:
    return registry.provide_db_repository(repository_type=param.type.func)


def provide_sync_db_session() -> Session:
    return registry.provide_db_session


def provide_sync_db_config() -> SQLAlchemySyncConfigT:
    return registry.provide_sync_config


def provide_async_db_config() -> SQLAlchemyAsyncConfigT:
    return registry.provide_async_config


def provide_importer_service(param: typer.CallbackParam) -> ImporterT:
    """Provide an importer service based on the type specified in the callback parameter."""
    importer_type = param.type.func
    return registry.provide_importer_service(importer_type=importer_type)


def provide_scheduler_service(param: typer.CallbackParam) -> SyncServiceT:
    """Provide a scheduling service based on the type specified in the callback parameter."""
    scheduler_type = param.type.func
    return registry.provide_scheduler_service(scheduler_type=scheduler_type)



from __future__ import annotations

from collections import deque
from datetime import date, datetime, timedelta
from itertools import chain
from random import shuffle
from typing import TYPE_CHECKING, Generator, Iterable
from uuid import UUID

from attr import define, field

from leaguemanager import models as m
from leaguemanager import services as s
from leaguemanager.schema import Day, GameTime, Matchup, Month, Weekday, Year

if TYPE_CHECKING:
    from leaguemanager.dependency import LeagueManager

__all__ = ["RoundRobinSchedule"]


@define
class RoundRobinSchedule:
    """Service for generating and managing round-robin schedules."""

    league_manager_registry: LeagueManager | None = field(default=None)
    _season: m.Season = field(default=None)

    def generate_schedule(self, shuffle_order: bool = True, phase_identifier: str = "Matchday") -> None:
        """Generate the round-robin schedule."""
        if len(self.teams) < 2:
            raise ValueError("At least two teams are required to generate a round-robin schedule.")

        # Calculate full number of rounds, and remaining phases (if any)
        rounds, remaining = divmod(self.ruleset.number_of_phases, self.ruleset.number_of_teams - 1)

        if rounds == 0:
            raise ValueError("Not enough phases to create a round-robin schedule.")

        phases = self.get_or_create_phases(phase_identifier=phase_identifier)
        season_matchups = self.create_season_matchups(shuffle_order, phases)
        fixtures = self.save_matchups_to_db(season_matchups)

        return fixtures

    def get_schedule(self) -> None:
        """Get the generated round-robin schedule."""
        # Implementation for retrieving the round-robin schedule
        raise NotImplementedError("Retrieving round-robin schedule not implemented.")

    def update_schedule(self) -> None:
        """Update the existing round-robin schedule."""
        # Implementation for updating the round-robin schedule
        raise NotImplementedError("Updating round-robin schedule not implemented.")

    def get_or_create_phases(self, phase_identifier: str = "Matchday") -> None:
        """Creates phases if they do not exist in the database."""

        if not self.ruleset.number_of_phases:
            raise ValueError("Number of phases is not set in the ruleset.")

        if not self.season.phases:
            phases = []
            for phase_number in range(1, self.ruleset.number_of_phases + 1):
                phase = m.Phase(
                    label=f"{phase_identifier} {phase_number}",
                    season_id=self.season.id,
                )
                phase = self.phase_service.create(phase, auto_commit=True)
                phases.append(phase)
        return self.season.phases

    def save_matchups_to_db(
        self,
        matchups: Iterable[Matchup],
    ) -> list[m.Fixture]:
        """Save matchups to the database as Fixtures and TeamStats."""
        if not self.league_manager_registry:
            raise ValueError("LeagueManager registry is not set.")
        fixtures = []
        for matchup in matchups:
            fixture = m.Fixture(
                season_id=self.season.id,
                phase_id=matchup.phase_id,
                label=f"{matchup.home_team.name} vs {matchup.away_team.name}",
                site_venue=f"Venue {matchup.venue_number}",
            )
            team_stat_a = m.TeamStats(
                season_id=self.season.id,
                fixture_id=fixture.id,
                label=f"{matchup.home_team.name} Stats",
            )
            team_stat_b = m.TeamStats(
                season_id=self.season.id,
                fixture_id=fixture.id,
                label=f"{matchup.away_team.name} Stats",
            )
            matchup.home_team.team_stats_id = team_stat_a.id
            matchup.away_team.team_stats_id = team_stat_b.id
            fixture = self.fixture_service.create(fixture, auto_commit=True)
            _ = self.team_service.update(matchup.home_team, auto_commit=True)
            _ = self.team_service.update(matchup.away_team, auto_commit=True)
            fixtures.append(fixture)

        return fixtures

    @property
    def registry(self) -> LeagueManager:
        """Get the LeagueManager registry."""
        if self.league_manager_registry is None:
            raise ValueError("LeagueManager registry is not set.")
        return self.league_manager_registry

    @property
    def season_service(self) -> s.SeasonService:
        """Get the SeasonService from the LeagueManager registry."""
        return self.registry.provide_db_service(service_type=s.SeasonService)

    @property
    def ruleset_service(self) -> s.RulesetService:
        """Get the RulesetService from the LeagueManager registry."""
        return self.registry.provide_db_service(service_type=s.RulesetService)

    @property
    def phase_service(self) -> s.PhaseService:
        """Get the PhaseService from the LeagueManager registry."""
        return self.registry.provide_db_service(service_type=s.PhaseService)

    @property
    def fixture_service(self) -> s.FixtureService:
        """Get the FixtureService from the LeagueManager registry."""
        return self.registry.provide_db_service(service_type=s.FixtureService)

    @property
    def team_stats_service(self) -> s.TeamStatsService:
        """Get the TeamStatsService from the LeagueManager registry."""
        return self.registry.provide_db_service(service_type=s.TeamStatsService)

    @property
    def team_service(self) -> s.TeamService:
        """Get the TeamService from the LeagueManager registry."""
        return self.registry.provide_db_service(service_type=s.TeamService)

    @property
    def season(self) -> m.Season:
        """Get the current season."""
        try:
            return self.season_service.get(self._season.id)
        except ValueError as e:
            raise ValueError(f"Season with ID {self._season.id} does not exist in the registry.") from e

    @property
    def ruleset(self) -> m.Ruleset:
        """Get the current ruleset."""
        if _ruleset := self.season.ruleset:
            return _ruleset
        raise ValueError(
            f"Ruleset for season {self.season.name} (ID: {self.season.id}) does not exist in the registry."
        )

    @property
    def teams(self) -> list[m.Team]:
        """Get the teams participating in the current season."""
        if _teams := self.season_service.all_teams(self.season.id):
            return _teams
        return [m.Team(name=f"Team {i + 1}") for i in range(self.ruleset.number_of_teams)]

    def create_season_matchups(self, shuffle_order: bool = True, phases: list[m.Phase] = None) -> list[Matchup]:
        """Generate matchups for the entire season."""

        if shuffle_order:
            shuffle(self.teams)

        all_matchups = []
        for phase, count in enumerate(phases, start=1):
            round_number = (count - 1) // (self.ruleset.number_of_teams - 1) + 1
            phase_matchups = self.phase_matchups(phase, count, round_number)
            all_matchups.extend(phase_matchups)

        return all_matchups

    def phase_matchups(self, phase: m.Phase, count: int, round_number: int) -> list[Matchup]:
        """Create events/matchups for a specific phase.

        See: https://en.wikipedia.org/wiki/Round-robin_tournament#Circle_method for
        details on round-robin scheduling. Splitting teams creates "bye week" team to
        ensure even matchups, and the phase number is used to rotate the teams for
        each phase of the round-robin schedule.
        """

        if len(self.teams) < 2:
            raise ValueError("At least two teams are required to create matchups.")

        home, away = self.split_teams(self.teams)

        # Create matchup by zipping home and away teams
        _matchups = list(zip(home, away, strict=False))

        # Anchor one team and rotate the rest based on phase
        if not count == 1:
            _teams = deque(chain(*_matchups))
            anchor = _teams.popleft()
            _teams.rotate(count)
            _teams.appendleft(anchor)
            home, away = self.split_teams([*_teams])
            _matchups = list(zip(home, away, strict=False))

        # Create Matchup objects for each matchup
        matchups = self.create_matchups(_matchups, phase, count, round_number)

        return matchups

    def create_matchups(
        self,
        matchups: Iterable[tuple[m.Team, m.Team]],
        phase: m.Phase,
        count: int,
        round_number: int,
    ) -> Generator[Matchup, None, None]:
        """Create Matchup objects from the list of team matchups."""
        for match_number, matchup_teams in enumerate(matchups, start=1):
            if round_number % 2 == 0:
                home, away = matchup_teams
            else:
                home, away = matchup_teams[::-1]

            yield Matchup(
                home_team=home,
                away_team=away,
                phase_id=phase.id,
                venue_number=self.generate_venue_number(match_number),
                game_time=self.generate_game_time(match_number, count, round_number),
            )

    def split_teams(self, team_obj: Iterable[m.Team]) -> tuple[list[m.Team], list[m.Team]]:
        """Split a list of teams in half, creating "bye week" team if number of teams is odd."""

        if len(team_obj) % 2 != 0:
            self.teams.append(m.Team(name="_bye"))

        return (
            team_obj[: len(team_obj) // 2],
            team_obj[len(team_obj) // 2 :],
        )

    def generate_venue_number(self, match_number: int) -> int:
        """Generates a venue number based on match_number and venue count set in the Ruleset.

        If there is only one venue, it always returns 1. If there are 3 venues and the match number is 1, it returns 1. If the match number is 2, it returns 2.
        If the match number is 3, it returns 3. If the match number is 4, it wraps around and returns 1.
        """

        venue_number = match_number % self.ruleset.venue_count
        return venue_number if venue_number != 0 else self.ruleset.venue_count

    def generate_game_time(self, match_number: int, phase: int) -> GameTime:
        """Generates a GameTime object based on when the match should be played.

        The date can be derived by incrementing the start date of the season by the product of
        the phase and number of days to increment between phases (i.e., 7 days for weekly matches).
        The match start time is determined by how many matches can be played concurrently (venue_count),
        the match number for that phase, as well as the game length and time between matches set in the Ruleset.
        """

        increment_by = self._increment_days()
        start_date = self.ruleset.start_date

        if not start_date:
            raise ValueError("Season start date is not set in the ruleset.")
        if not isinstance(start_date, (datetime, date)):
            raise TypeError("Season start date must be a datetime or date object.")

        # Note: This currently works for weekly scheduling only.
        match_date = start_date + timedelta(days=increment_by[0] * phase)
        game_time = self._increment_time(match_number, match_date)
        return game_time

    def _list_of_days(self) -> list[str]:
        """Return a list of days when fixtures can be scheduled based on the ruleset."""
        days = []
        if self.ruleset.mon_fixtures:
            days.append(Weekday.MONDAY)
        if self.ruleset.tue_fixtures:
            days.append(Weekday.TUESDAY)
        if self.ruleset.wed_fixtures:
            days.append(Weekday.WEDNESDAY)
        if self.ruleset.thu_fixtures:
            days.append(Weekday.THURSDAY)
        if self.ruleset.fri_fixtures:
            days.append(Weekday.FRIDAY)
        if self.ruleset.sat_fixtures:
            days.append(Weekday.SATURDAY)
        if self.ruleset.sun_fixtures:
            days.append(Weekday.SUNDAY)
        return days

    def _increment_days(self) -> list[int]:
        """Return a list of days to increment based on the ruleset."""

        days = self._list_of_days()
        if not days:
            raise ValueError("No days available for scheduling fixtures based on the ruleset.")
        if len(days) >= 1:
            raise NotImplementedError("Currently only supports weekly scheduling.")
        return [7]  # Default to weekly increment

    def _increment_time(self, match_count: int, match_date: datetime) -> GameTime:
        """Return the time increment between matches based on the ruleset."""

        if self.ruleset.game_length is None or self.ruleset.time_between is None:
            raise ValueError("Game length and time between matches must be set in the ruleset.")

        increment = timedelta(minutes=self.ruleset.game_length + self.ruleset.time_between)

        if match_count <= self.ruleset.venue_count:
            hour, minute = match_date.hour, match_date.minute
        else:
            # convert time delta into hours, minutes
            multiplier, _ = divmod((match_count), self.ruleset.venue_count)
            if match_count % self.ruleset.venue_count == 0:
                multiplier -= 1
            hour, minute = self._convert_to_hours_and_minutes(increment, multiplier)
            hour += match_date.hour

        return GameTime(
            month=Month(match_date.month),
            day=Day(match_date.day),
            year=Year(match_date.year),
            hour=hour,
            minute=minute,
            total_duration_in_minutes=self.ruleset.game_length + self.ruleset.time_between,
        )

    def _convert_to_hours_and_minutes(self, td: timedelta, multiplier: int) -> tuple[int, int]:
        """Convert a timedelta to (hours, minutes)."""

        _minutes = int(td.total_seconds() // 60)
        total_minutes = _minutes * multiplier
        hours, minutes = divmod(total_minutes, 60)
        return hours, minutes



from .bracket import BracketSchedule
from .round_robin import RoundRobinSchedule
from .round_robin_playoff import RoundRobinPlayoffSchedule
from .tournament import TournamentSchedule

__all__ = [
    "RoundRobinSchedule",
    "RoundRobinPlayoffSchedule",
    "TournamentSchedule",
    "BracketSchedule",
]



from __future__ import annotations

from typing import TYPE_CHECKING

from attrs import define, field

from leaguemanager import models as m

if TYPE_CHECKING:
    from .safe_date import GameTime


@define
class Matchup:
    """Represents a matchup between two teams in a league.

    Attributes:
        home_team (Team): The home team in the matchup.
        away_team (Team): The away team in the matchup.
        phase_id (int): The phase id of the matchup.
        venue_number (int): The venue number for the matchup.
        game_time (GameTime): The scheduled game time for the matchup, if available.
    """

    home_team: m.Team
    away_team: m.Team
    phase_id: int = field(default=1)
    venue_number: int = field(default=1)
    game_time: GameTime | None = field(default=None)



from typing import Protocol, runtime_checkable


@runtime_checkable
class ScheduleServiceBase(Protocol):
    """Protocol for scheduling services.

    This protocol defines the methods that a scheduling service should implement.
    """

    def generate_schedule(self) -> None:
        """Generate the schedule for the competition."""
        raise NotImplementedError("This method should be implemented by subclasses.")

    def get_schedule(self) -> None:
        """Get the generated schedule."""
        raise NotImplementedError("This method should be implemented by subclasses.")

    def update_schedule(self) -> None:
        """Update the existing schedule."""
        raise NotImplementedError("This method should be implemented by subclasses.")



from datetime import date
from enum import Enum
from typing import Self
from uuid import UUID

from attrs import define, field, validators

__all__ = ["Year", "Month", "Day", "GameTime"]


@define(frozen=True)
class Year:
    year: int

    def __attrs_post_init__(self):
        if not (1999 <= self.year <= 2099):
            raise ValueError(f"Year {self.year} is out of valid range (1999-2099).")


@define(frozen=True)
class MonthValue:
    number: int
    days: int  # Number of days in the month

    def __attrs_post_init__(self) -> None:
        if not 1 <= self.number <= 12:
            raise ValueError(f"Invalid month number: {self.number}")
        if not 1 <= self.days <= 31:
            raise ValueError(f"Invalid days in month: {self.days}")


class Weekday(str, Enum):
    MONDAY = "Monday"
    TUESDAY = "Tuesday"
    WEDNESDAY = "Wednesday"
    THURSDAY = "Thursday"
    FRIDAY = "Friday"
    SATURDAY = "Saturday"
    SUNDAY = "Sunday"

    def __str__(self) -> str:
        return self.value

    @classmethod
    def _missing_(cls, value: object) -> Self:
        if not isinstance(value, str):
            raise TypeError("Expected str")
        for day in cls:
            if day.value.lower() == value.lower():
                return day
        raise ValueError(f"No such weekday: {value}")


class Month(Enum):
    JANUARY = MonthValue(1, 31)
    FEBRUARY = MonthValue(2, 28)  # Note: February can have 29 days in leap years
    # For simplicity, assume February has 28 days here.
    # Leap year handling can be added later if needed.
    MARCH = MonthValue(3, 31)
    APRIL = MonthValue(4, 30)
    MAY = MonthValue(5, 31)
    JUNE = MonthValue(6, 30)
    JULY = MonthValue(7, 31)
    AUGUST = MonthValue(8, 31)
    SEPTEMBER = MonthValue(9, 30)
    OCTOBER = MonthValue(10, 31)
    NOVEMBER = MonthValue(11, 30)
    DECEMBER = MonthValue(12, 31)

    def __str__(self) -> str:
        return self.name

    def __int__(self) -> int:
        return self.value.number

    def valid_day(self, day: int) -> None:
        if not 1 <= day <= self.value.days:
            raise ValueError(f"Invalid day {day} for {self}")

    @classmethod
    def _missing_(cls, value: object) -> Self:
        if not isinstance(value, int):
            raise TypeError("Expected int")
        for m in cls:
            if m.value.number == value:
                return m
        raise ValueError(f"No such month: {value}")


@define(frozen=True)
class Day:
    value: int

    def __attrs_post_init__(self) -> None:
        if self.value <= 0:
            raise ValueError(f"Invalid day: {self.value}")

    @classmethod
    def of(cls, month: Month, day: int) -> Self:
        # Ensure day is within Month's range:
        month.valid_day(day)
        return cls(day)


@define
class GameTime:
    month: Month
    day: Day
    year: Year
    hour: int = field(default=0, validator=[validators.lt(25), validators.ge(0)])
    minute: int = field(default=0, validator=[validators.lt(61), validators.ge(0)])
    total_duration_in_minutes: int = field(default=0, validator=[validators.ge(0)])

    def __str__(self) -> str:
        return f"{self.hour:02}:{self.minute:02}"



from __future__ import annotations

from typing import TYPE_CHECKING, TypeVar

from advanced_alchemy.config import SQLAlchemyAsyncConfig, SQLAlchemySyncConfig
from advanced_alchemy.repository import SQLAlchemyAsyncRepository, SQLAlchemySyncRepository
from advanced_alchemy.service import SQLAlchemyAsyncRepositoryService, SQLAlchemySyncRepositoryService

from leaguemanager.models.base import UUIDBase
from leaguemanager.services.scheduling import (
    BracketSchedule,
    RoundRobinPlayoffSchedule,
    RoundRobinSchedule,
    TournamentSchedule,
)
from leaguemanager.services.template_loader.league_importer import Importer

ModelT = TypeVar("ModelT", bound=UUIDBase)
SyncRepositoryT = TypeVar("RepositoryT", bound=SQLAlchemySyncRepository)
AsyncRepositoryT = TypeVar("RepositoryT", bound=SQLAlchemyAsyncRepository)
SyncServiceT = TypeVar("ServiceT", bound=SQLAlchemySyncRepositoryService[ModelT])
AsyncServiceT = TypeVar("ServiceT", bound=SQLAlchemyAsyncRepositoryService[ModelT])
SQLAlchemySyncConfigT = TypeVar("ConfigT", bound=SQLAlchemySyncConfig)
SQLAlchemyAsyncConfigT = TypeVar("ConfigT", bound=SQLAlchemyAsyncConfig)
ImporterT = TypeVar("ImporterT", bound=Importer)
ScheduleServiceT = TypeVar(
    "ScheduleServiceT", bound=BracketSchedule | RoundRobinSchedule | RoundRobinPlayoffSchedule | TournamentSchedule
)



from __future__ import annotations

from typing import TYPE_CHECKING

from attr import define, field

from leaguemanager import models as m

if TYPE_CHECKING:
    from leaguemanager.services import FixtureService, RulesetService, SeasonService

__all__ = ["TournamentSchedule"]


@define
class TournamentSchedule:
    """Service for generating and managing tournament playoff schedules."""

    season_service: SeasonService = field(default=None)
    ruleset_service: RulesetService = field(default=None)
    fixture_service: FixtureService = field(default=None)

    def generate_schedule(self) -> None:
        """Generate the tournament playoff schedule."""
        # Implementation for generating a tournament playoff schedule
        raise NotImplementedError("Tournament playoff schedule generation not implemented.")

    def get_schedule(self) -> None:
        """Get the generated tournament playoff schedule."""
        # Implementation for retrieving the tournament playoff schedule
        raise NotImplementedError("Retrieving tournament playoff schedule not implemented.")

    def update_schedule(self) -> None:
        """Update the existing tournament playoff schedule."""
        # Implementation for updating the tournament playoff schedule
        raise NotImplementedError("Updating tournament playoff schedule not implemented.")



# from leaguemanager.services.template_loader import CSVLoader, ExcelLoader, GoogleSheetsLoader, JSONLoader, MemoryLoader

# __all__ = [
#     "CSVLoader",
#     "ExcelLoader",
#     "GoogleSheetsLoader",
#     "JSONLoader",
#     "MemoryLoader",
# ]
